#ifndef MESSAGINGUTILS_H_
#define MESSAGINGUTILS_H_

#include <QObject>
#include <QDomDocument>
#include <QDateTime>
#include <QStringList>

//Namespace definitions
#define NS_XSI "http://www.w3.org/2001/XMLSchema-instance"
#define NS_XSD "http://www.w3.org/2001/XMLSchema"
#define NS_SOAP "http://schemas.xmlsoap.org/soap/envelope/"
#define NS_T "http://schemas.microsoft.com/exchange/services/2006/types"
#define NS_MSG "http://schemas.microsoft.com/exchange/services/2006/messages"

//Used for http request header to define web service operation
#define ACTION_URL "http://schemas.microsoft.com/exchange/services/2006/messages/"

//Set MessagingUtils Debug on/off
#define MU_DEBUG
//#undef MU_DEBUG

class Meeting;
class Room;

//All supported operations for web services
//Operations are described here: http://msdn.microsoft.com/en-us/library/bb409286.aspx
enum RequestCommandId
{
	ReqCmdNoCommand = 0,
	ReqCmdAddDelegate, //1
	ReqCmdConvertId, //2 Used for converting fetched IDs to match the format in GetItem request
	ReqCmdCopyFolder, //3
	ReqCmdCopyItem, //4
	ReqCmdCreateAttachment, //5
	ReqCmdCreateFolder, //6
	ReqCmdCreateItem, //7
	ReqCmdCreateManagedFolder, //8
	ReqCmdDeleteAttachment, //9
	ReqCmdDeleteFolder, //10
	ReqCmdDeleteItem, //11
	ReqCmdExpandDL, //12
	ReqCmdFindFolder, //13
	ReqCmdFindItem, //14
	ReqCmdGetAttachment, //15
	ReqCmdGetDelegate, //16
	ReqCmdGetEvents, //17
	ReqCmdGetFolder, //18
	ReqCmdGetItem, //19 Used for getting the detailed information about some item (calendar event)
	ReqCmdGetUserAvailability, //20 Used for fetching meetings
	ReqCmdGetUserOofSettings, //21
	ReqCmdMoveFolder, //22
	ReqCmdMoveItem, //23
	ReqCmdRemoveDelegate, //24
	ReqCmdResolveNames, //25
	ReqCmdSendItem, //26
	ReqCmdSetUserOofSettings, //27
	ReqCmdSubscribe, //28
	ReqCmdSyncFolderHierarchy, //29
	ReqCmdSyncFolderItems, //30
	ReqCmdUnsubscribe, //31
	ReqCmdUpdateDelegate, //32
	ReqCmdUpdateFolder, //33
	ReqCmdUpdateItem //34
};

//TODO: Define more specific errors
enum MessagingError
{
	MsgErrNoError = 0,
	MsgErrSomeError
};

//! Struct to present a node in SOAP envelope message.
/*!
 * 	Struct to present a node in SOAP envelope message.
 *  Request messages are formed by using predefined arrays
 *  of MessageBodyElement objects.
 */
typedef struct MessageBodyElement
{
	//! Depth of the node in document tree. For example, root node should be 0.
	int iTraversalLevel;
	//! Namespace of the node. QString::null if not specified.
	QString iNamespace;
	//! Name of the node.
	QString iElementName;
	//! Type of the node. Currently QDomNode::ElementNode and QDomNode::AttributeNode are supported.
	QDomNode::NodeType iNodeType;
};

//SOAP Envelope templates for web service operations
//Operations are described here: http://msdn.microsoft.com/en-us/library/bb409286.aspx
/*
 * NOTES: 
 * - First node's TraversalLevel must always be 0.
 * - Namespace can be defined as an URI or QString::null
 * - NodeType can be QDomNode::ElementNode or QDomNode::AttributeNode
 * - Nodes must be defined in descending order, according to schema.
 * - Attribute must be defined right after it's owner element and it's traversal level must be same.
 */
const MessageBodyElement reqCmdArrayEnvelopeBase[] =
{
		{ 0, NS_SOAP, "Envelope", QDomNode::ElementNode }/*,
		{ 0, "xmlns", "xsi", QDomNode::AttributeNode },
		{ 0, "xmlns", "xsd", QDomNode::AttributeNode },
		{ 0, "xmlns", "soap", QDomNode::AttributeNode },
		{ 0, "xmlns", "t", QDomNode::AttributeNode },*/
};

const MessageBodyElement reqCmdMailboxElement[] =
{
		{ 0, NS_T, "MailboxData", QDomNode::ElementNode },
		{ 1, NS_T, "Email", QDomNode::ElementNode },
		{ 2, NS_T, "Address", QDomNode::ElementNode },
		{ 1, NS_T, "AttendeeType", QDomNode::ElementNode },
		{ 1, NS_T, "ExcludeConflicts", QDomNode::ElementNode },
};

const MessageBodyElement reqCmdArrayGetUserAvailability[] =
{
		{ 0, NS_SOAP, "Body", QDomNode::ElementNode },
		{ 1, NS_MSG, "GetUserAvailabilityRequest", QDomNode::ElementNode },
		{ 2, NS_T, "TimeZone", QDomNode::ElementNode },
		{ 3, QString::null, "Bias", QDomNode::ElementNode },
		{ 3, QString::null, "StandardTime", QDomNode::ElementNode },
		{ 4, QString::null, "Bias", QDomNode::ElementNode },
		{ 4, QString::null, "Time", QDomNode::ElementNode },
		{ 4, QString::null, "DayOrder", QDomNode::ElementNode },
		{ 4, QString::null, "Month", QDomNode::ElementNode },
		{ 4, QString::null, "DayOfWeek", QDomNode::ElementNode },
		//</standardtime>
		{ 3, QString::null, "DaylightTime", QDomNode::ElementNode },
		{ 4, QString::null, "Bias", QDomNode::ElementNode },
		{ 4, QString::null, "Time", QDomNode::ElementNode },
		{ 4, QString::null, "DayOrder", QDomNode::ElementNode },
		{ 4, QString::null, "Month", QDomNode::ElementNode },
		{ 4, QString::null, "DayOfWeek", QDomNode::ElementNode },
		//</daylighttime>
		//</timezone>
		//Use meetingroom addresses to fetch meetings for rooms. Mailboxdata-element for each.
		{ 2, QString::null, "MailboxDataArray", QDomNode::ElementNode },

		//MAILBOX DATA(s) MUST BE ADDED AFTERWARDS
		//SEE: reqCmdMailboxElement[]

		//</mailboxdeataarray>
		{ 2, NS_T, "FreeBusyViewOptions", QDomNode::ElementNode },
		//Use Timewindow to define from which period we want to fetch meetings
		{ 3, NS_T, "TimeWindow", QDomNode::ElementNode },
		{ 4, NS_T, "StartTime", QDomNode::ElementNode },
		{ 4, NS_T, "EndTime", QDomNode::ElementNode },
		//</timewindow>
		{ 3, NS_T, "MergedFreeBusyIntervalInMinutes", QDomNode::ElementNode },
		{ 3, NS_T, "RequestedView", QDomNode::ElementNode },
		//</freebusyviewoptions>
		//</getuseravailabilityrequest>
};

const MessageBodyElement reqCmdArrayGetCalendarItem[] =
{
		{ 0, NS_SOAP, "Body", QDomNode::ElementNode },
		{ 1, NS_MSG, "GetItem", QDomNode::ElementNode },
		//{ 1, QString::null, "xmlns", QDomNode::AttributeNode },
		{ 2, QString::null, "ItemShape", QDomNode::ElementNode },
		{ 3, NS_T, "BaseShape", QDomNode::ElementNode },
		{ 2, QString::null, "ItemIds", QDomNode::ElementNode },
		{ 3, NS_T, "ItemId", QDomNode::ElementNode },
		{ 3, QString::null, "Id", QDomNode::AttributeNode }
};

const MessageBodyElement reqCmdArrayConvertIdHeader[] =
{
		{ 0, NS_SOAP, "Header", QDomNode::ElementNode },
		{ 1, NS_T, "RequestServerVersion", QDomNode::ElementNode },
		{ 1, QString::null, "Version", QDomNode::AttributeNode },

};

const MessageBodyElement reqCmdArrayConvertId[] =
{
		{ 0, NS_SOAP, "Body", QDomNode::ElementNode },
		{ 1, NS_MSG, "ConvertId", QDomNode::ElementNode },
		//{ 1, QString::null, "xmlns:t", QDomNode::AttributeNode },
		//{ 2, "xmlns", "t", QDomNode::AttributeNode },
		{ 1, QString::null, "DestinationFormat", QDomNode::AttributeNode },
		{ 2, QString::null, "SourceIds", QDomNode::ElementNode },
		{ 3, NS_T, "AlternateId", QDomNode::ElementNode },
		{ 3, QString::null, "Format", QDomNode::AttributeNode },
		{ 3, QString::null, "Id", QDomNode::AttributeNode },
		{ 3, QString::null, "Mailbox", QDomNode::AttributeNode },
};


//! Base class for Request and Response message classes
/*!
 *  This class provides basic functionality to use generated SOAP envelope DomDocument object.
 */
class BaseMessage : public QObject
{
	Q_OBJECT

public:

	//! Constructor.
	BaseMessage();
	//! Destructor.
	virtual ~BaseMessage();

	//! Current generated SOAP envelope DomDocument object as byte array.
	/*!
	 *	Returns current message envelope as byte array.
	 *  This array is used to provide a content for http request.
	 * \return current message envelope as byte array.
	 */ 
	QByteArray getMessage();

protected:

	//! List of QDomNode objects matching desired criteria.
	/*!
	 *  Returns a list of nodes that matches the given aNodeName.
	 *  Optional parameters can be provided to narrow the search.
	 * \param aNodeName Name of the node(s).
	 * \param aParentName Name of the parent node. Optional. 
	 *  If not specified, base node of the QDomDocument is used.
	 * \param aRootNode Root node for the search. Optional.
	 *  If provided, only it's childnodes will be searched.
	 *  If not specified, base node of the QDomDocument is used.
	 * \return List of QDomNodes matching the criteria.
	 */
	QList<QDomNode> getNodesByName( const QString& aNodeName,
									const QString& aParentName = QString::null,
									QDomNode* aRootNode = NULL );

	//! Single node matching desired criteria.
	/*!
	 *  Returns a single node that matches the given aNodeName.
	 *  Optional parameters can be provided to narrow the search.
	 * \param aNodeName Name of the node.
	 * \param aNodeType Type of the node. Currently QDomNode::ElementNode and QDomNode::AttributeNode are supported.
	 * \param aParentName Name of the parent node. Optional. 
	 *  If not specified, base node of the QDomDocument is used.
	 * \param aIndex If multiple results are found, an index of desired node can be used. Optional.
	 * \param aRootNode Root node for the search. Optional.
	 *  If provided, only it's childnodes will be searched.
	 *  If not specified, base node of the QDomDocument is used.
	 * \return QDomNode matching the criteria.
	 */
	QDomNode getNodeFromDocument( const QString& aNodeName,
									QDomNode::NodeType aNodeType,
									const QString& aParentName = QString::null,
									int aIndex = 0,
									QDomNode* aRootNode = NULL );

	//! Single element matching desired criteria.
	/*!
	 *  See getNodeFromDocument for details.
	 */
	QDomNode getElementFromDocument( const QString& aElementName,
									const QString& aParentName = QString::null,
									int aIndex = 0,
									QDomNode* aRootNode = NULL );
	
	//! Single attribute matching desired criteria.
	/*!
	 *  See getNodeFromDocument for details.
	 */
	QDomNode getAttributeFromDocument( const QString& aAttributeName,
									const QString& aParentName = QString::null,
									int aIndex = 0,
									QDomNode* aRootNode = NULL );

	//! Compare node name to string.
	/*!
	 * 	Checks if name of the provided node matches aName.
	 *  Node can be QDomNode::ElementNode or QDomNode::AttributeNode.
	 * \param aNode Node to compare.
	 * \param aName Name to compare.
	 * \return True if comparison matches. Otherwise false.
	 *  NOTE: This function does comparison for node's local name, so it ignores possible prefixes in a tag name.
	 */
	bool matchName( const QDomNode& aNode, const QString& aName  );



protected:

	/*!
	 * Current generated SOAP envelope DomDocument object.
	 */
	QDomDocument* iMessage;

};

//! Base class for Requests
/*!
 *  This class provides basic functionality to use generated SOAP envelope DomDocument object for http requests.
 */
class RequestMessage : public BaseMessage
{

public:

	//! Constructor
	/*!
	 * \param aCommandId Id to identify operation type.
	 */
	RequestMessage( RequestCommandId aCommandId = ReqCmdNoCommand );
	RequestMessage( const QString& aFileName );
	//! Destructor
	virtual ~RequestMessage(){};

public:

	//! Add node to message
	/*!
	 *  Adds node to a request message. Parameters can be provided to specify location.
	 * \param aNode Node to add.
	 * \param aNodeType Type of node. QDomNode::ElementNode and QDomNode::AttributeNode are supported. Default is QDomNode::ElementNode.
	 * \param aParentName Name of the parent. Optional. If provided, node will be added as it's child. Otherwise root of the QDomDocument is used.
	 * \param aIndex Index to add. Optional. If multiple parents are found, index to add can be specified.
	 * \param aRootNode Root node to search. Optional. Searches only it's child nodes for a parent if provided. Otherwise root of the QDomDocument is used.
	 * \return An error code.
	 */
	int addNode( const QDomNode& aNode,
					QDomNode::NodeType aNodeType = QDomNode::ElementNode,
					const QString& aParentName = QString::null,
					int aIndex = 0,
					QDomNode* aRootNode = NULL );

	//! Set value of the node
	/*!
	 *  Sets the value of specified node. Parameters can be used to narrow the search.
	 * \param aNodeName Name of the node.
	 * \param aValue Value to set.
	 * \param aNodeType Type of node. QDomNode::ElementNode and QDomNode::AttributeNode are supported. Default is QDomNode::ElementNode.
	 * \param aParentName Name of the parent node. Optional. 
	 *  If not specified, base node of the QDomDocument is used.
	 * \param aIndex If multiple results are found, an index of desired node can be used. Optional.
	 * \param aRootNode Root node for the search. Optional.
	 *  If provided, only it's childnodes will be searched.
	 *  If not specified, base node of the QDomDocument is used.
	 * \return An error code.
	 */
	int setNodeValue( const QString& aNodeName,
						const QString& aValue,
						QDomNode::NodeType aNodeType = QDomNode::ElementNode,
						const QString& aParentName = QString::null,
						int aIndex = 0,
						QDomNode* aRootNode = NULL );

	//! Create the structure of message (soap:Body)
	/*
	 * \return An error code.
	 */
	int createMessageStructure();
	//! Create the structure of message (soap:Body)
	/*
	 * \param aCommandId Id of command. Structure is constructed according to this.
	 * \return An error code.
	 */
	int createMessageStructure( RequestCommandId aCommandId );
	
	//! Create the base of the request envelope (soap:Envelope)
	/*!
	 * \return An error code.
	 */
	int createEnvelopeBase();

	//! Get operation specific content type string for http request header.
	/*!
	 * \param aCommandId Id of the command (operation)
	 * \return Content type string for header
	 */
	QString getContentTypeForHeader( RequestCommandId aCommandId = ReqCmdNoCommand );

protected:

	//! Construct an array of MessageBodyElements to a node.
	/*!
	 *  Constructs an array of predefined MessageBodyElements to node.
	 *  This function is used to form different parts of message from MessageBodyElements to QDomNodes
	 * \param aArray Array of MessageBodyElements.
	 * \param aSize Size of an array.
	 * \return Created node.
	 */
	QDomNode constructArrayToNode( const MessageBodyElement* aArray, int aSize );
	
protected:

	//! Id of current operation
	RequestCommandId iCurrCmd;
};

//! Base class for Responses.
/*!
 *  This class provides basic functionality to use generated SOAP envelope DomDocument object from http responses.
 */
class ResponseMessage : public BaseMessage
{

public:

	//! Constructor
	ResponseMessage();

	//! Constructor
	/*!
	 * \param aData Byte array to construct ResponseMessage from.
	 */
	ResponseMessage( const QByteArray& aData );

	//! Destructor
	virtual ~ResponseMessage(){};

	//! Get the value of node.
	/*!
	 *  Gets a value of single node. Optional parameters are used to define desired node in more detail.
	 * \param aNodeName Name of the node.
	 * \param aNodeType Type of the node. QDomNode::ElementNode and QDomNode::AttributeNode are supported.
	 * \param aParentname Name of the parent node. Optional. Can be specified to narrow the search.
	 * \param aIndex Index of the node. Optional. If multiple results are found, index can be defined to specify the result.
	 * \param aRootNode Root node. Optional. If specified, only it's children are included in search. Otherwise, root node of QDomDocument is used.
	 * \return Value of the node.
	 */
	QString getNodeValue( const QString& aNodeName,
							QDomNode::NodeType aNodeType,
							const QString& aParentName = QString::null,
							int aIndex = 0,
							QDomNode* aRootNode = NULL );

	//! Checks if there are any errors in response.
	/*!
	 *  If any value of status elements in response message is not success, this function
	 *  returns true.
	 * \return True if errors exist. Otherwise false.
	 */
	bool hasErrors();

protected:

};


//Operation specific inherited classes


//! Request message for GetUserAvailability operation.
class ReqMsgGetUserAvailability : public RequestMessage
{

	enum UserAttendeeType
	{
		AttendeeOrganizer = 0,
		AttendeeRequired,
		AttendeeOptional,
		AttendeeRoom,
		AttendeeResource
	};

public:

	//! Constructor
	ReqMsgGetUserAvailability();
	//! Destructor
	virtual ~ReqMsgGetUserAvailability(){};
	//! Set the current time zone.
	/*
	 * \return An error code.
	 */
	int setTimeZone(); //TODO: initializable with parameters?
	
	//! Add user (mailbox) to search availability status from.
	/*
	 * Add user (mailbox) to search calendar events from. 
	 * Multiple users can be added in one request.
	 * \param aAddress User's EMail address.
	 * \param aAttendeeType Type of the attendance. Default is "Required".
	 * \param aExcludeConflicts Exclude conflicting meetings. Default is "false".
	 * \return An error code.
	 */
	int addUser( const QString& aAddress, const QString& aAttendeeType = "Required", const QString& aExcludeConflicts = "false" ); //TODO: Is resource a correct type for room?
	
	//! Set the time window for request.
	/*!
	 * Set the time window to search availability statuses from.
	 * \param aStart Start date time.
	 * \param aEnd End date time.
	 * \return An error code.
	 * NOTE: Server might give an error if too wide time window is specified.
	 * ~3 weeks should be safe.
	 */
	int setTimeWindow( const QDateTime& aStart, const QDateTime& aEnd );

private:

};

//! Request message for ConvertId operation.
/*!
 * This is used to convert HexEntryId type unique Id's of calendar events
 * To EwsLegacyId type. Latter type of Id must be provided in order to get detailed information for calendar event.
 */
class ReqMsgConvertMeetingId : public RequestMessage
{
public:
	//! Constructor
	/*!
	 * \param aItemId HexEntryId type Id to convert. Optional (though must be set later).
	 * \param aMailBox EMail address of calendar event's owner.
	 */
	ReqMsgConvertMeetingId( const QString& aItemId = QString::null, const QString& aMailbox = QString::null );
	
	//! Destructor
	virtual ~ReqMsgConvertMeetingId(){};
	
	//! Set item id.
	/*
	 * \param aItemId Id of calendar event in HexEntryId format
	 * \return An error code.
	 */
	int setItemId( const QString& aItemId );
	
	//! Set Mailbox.
	/*!
	 *  Set the calendar event owner's mailbox EMail address.
	 * \param aMailBox Mailbox address.
	 * \return An error code.
	 */
	int setMailbox( const QString& aMailbox );
	
private:
	
};

//! Request message for GetItem (Calendar) operation.
/*!
 *  This is used to get more detailed information about calendar event.
 */
class ReqMsgGetCalendarItem : public RequestMessage
{
public:

	//! Constructor
	/*!
	 * \param aItemId secondary Id ( EwsLegacyId type obtained from ConvertId operation ) of calendar event.
	 */
	ReqMsgGetCalendarItem( const QString& aItemId = QString::null );
	
	//! Destructor
	virtual ~ReqMsgGetCalendarItem(){};
	
	//! Set the Id of calendar event.
	/*!
	 * This must be EwsLegacyId type obtained from ConvertId operation.
	 * \param aItemId Id of item.
	 * \return An error code.
	 */
	int setItemId( const QString& aItemId );	
	//int setChangeKey( const QString& aChangeKey ){};
	
private:

};

//! Response message for GetUserAvailability operation.
class ResMsgGetUserAvailability : public ResponseMessage
{
public:

	//! Constructor
	/*!
	 * \param aData Byte array to construct response message from.
	 */
	ResMsgGetUserAvailability( const QByteArray& aData );
	
	//! Destructor
	virtual ~ResMsgGetUserAvailability(){};
	//! Get Meetings from constructed response.
	/*!
	 * \param aList List of Meeting objects to fill.
	 * \param aRoom Room for Meeting objects.
	 * \return An error code.
	 */
	int getMeetingsFromResponse( QList<Meeting*>& aList, const Room &aRoom );
};

//! Response message for GetItem (Calendar) operation.
class ResMsgGetCalendarItem : public ResponseMessage
{
public:
	//! Constructor
	/*!
	 * \param aData Byte array to construct response message from.
	 */
	ResMsgGetCalendarItem( const QByteArray& aData ) : ResponseMessage( aData ){};
	
	//! Destructor
	virtual ~ResMsgGetCalendarItem(){};
	
	//! Get meeting details from constructed response.
	/*!
	 * \param aMeeting Meeting to get details for.
	 * \return An error code.
	*/
	int getMeetingDetailsFromResponse( Meeting& aMeeting );

};

#endif /*MESSAGINGUTILS_H_*/
